#include <stdio.h>
#include <stdlib.h>
#include <phidget22.h>
#include "PhidgetHelperFunctions.h"


/**
* Configures the device's DataInterval and ChangeTrigger.
* Displays info about the attached phidget channel.
* Fired when a Phidget channel with onAttachHandler registered attaches
*
* @param ph The Phidget channel that fired the attach event
* @param *ctx Context pointer
*/
static void CCONV onAttachHandler(PhidgetHandle ph, void *ctx) {
	PhidgetReturnCode prc;
	int32_t serialNumber;
	PhidgetHandle hub;
	int32_t hubPort;
	int32_t channel;

	/*
	*	Set the DataInterval inside of the attach handler to initialize the device with this value.
	*	DataInterval defines the minimum time between VoltageChange events.
	*	DataInterval can be set to any value from MinDataInterval to MaxDataInterval.
	*/
	printf("\tSetting DataInterval to 1000ms\n");
	prc = PhidgetVoltageInput_setDataInterval((PhidgetVoltageInputHandle)ph, 1000);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set DataInterval: \n\t");
		DisplayError(prc);
		return;
	}

	/*
	*	Set the VoltageChangeTrigger inside of the attach handler to initialize the device with this value.
	*	VoltageChangeTrigger will affect the frequency of VoltageChange events, by limiting them to only occur when
	*	the voltage changes by at least the value set.
	*/
	printf("\tSetting Voltage ChangeTrigger to 0.0\n");
	prc = PhidgetVoltageInput_setVoltageChangeTrigger((PhidgetVoltageInputHandle)ph, 0.0);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set VoltageChangeTrigger: \n\t");
		DisplayError(prc);
		return;
	}

	prc = Phidget_getDeviceSerialNumber(ph, &serialNumber);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get DeviceSerialNumber: \n\t");
		DisplayError(prc);
		return;
	}

	prc = Phidget_getChannel(ph, &channel);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get Channel: \n\t");
		DisplayError(prc);
		return;
	}

	//Check if this is a VINT device
	prc = Phidget_getHub(ph, &hub);
	if (EPHIDGET_WRONGDEVICE != prc) {
		prc = Phidget_getHubPort(ph, &hubPort);
		if (EPHIDGET_OK != prc) {
			fprintf(stderr, "Runtime Error -> Get HubPort: \n\t");
			DisplayError(prc);
			return;
		}
		printf("\n[Attach Event]:\n\t-> Serial Number: %d\n\t-> Hub Port: %d\n\t-> Channel %d\n\n", serialNumber, hubPort, channel);
	}
	else {
		printf("\nAttach Event:\n\t-> Serial Number: %d\n\t-> Channel %d\n\n", serialNumber, channel);
	}
}

/**
* Displays info about the detached phidget channel.
* Fired when a Phidget channel with onDetachHandler registered detaches
*
* @param ph The Phidget channel that fired the detach event
* @param *ctx Context pointer
*/
static void CCONV onDetachHandler(PhidgetHandle ph, void *ctx) {
	PhidgetReturnCode prc;
	PhidgetHandle hub;
	int serialNumber;
	int hubPort;
	int channel;

	prc = Phidget_getDeviceSerialNumber(ph, &serialNumber);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get DeviceSerialNumber: \n\t");
		DisplayError(prc);
		return;
	}

	prc = Phidget_getChannel(ph, &channel);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get Channel: \n\t");
		DisplayError(prc);
		return;
	}

	//Check if this is a VINT device
	prc = Phidget_getHub(ph, &hub);
	if (EPHIDGET_WRONGDEVICE != prc) {
		prc = Phidget_getHubPort(ph, &hubPort);
		if (EPHIDGET_OK != prc) {
			fprintf(stderr, "Runtime Error -> Get HubPort: \n\t");
			DisplayError(prc);
			return;
		}
		printf("\nDetach Event:\n\t-> Serial Number: %d\n\t-> Hub Port: %d\n\t-> Channel %d\n\n", serialNumber, hubPort, channel);
	}
	else {
		printf("\nDetach Event:\n\t-> Serial Number: %d\n\t-> Channel %d\n\n", serialNumber, channel);
	}
}

/**
* Writes phidget error info to stderr.
* Fired when a Phidget channel with onErrorHandler registered encounters an error in the library
*
* @param ph The Phidget channel that fired the error event
* @param *ctx Context pointer
* @param errorCode the code associated with the error of enum type Phidget_ErrorEventCode
* @param *errorString string containing the description of the error fired
*/
static void CCONV onErrorHandler(PhidgetHandle phid, void *ctx, Phidget_ErrorEventCode errorCode, const char *errorString) {

	fprintf(stderr, "[Phidget Error Event] -> %s (%d)\n", errorString, errorCode);
}

/**
* Sets the event handlers for Phidget Attach, Phidget Detach, Phidget Error events
*
* @param ph The Phidget channel to add event handlers to
* @return 0 if the operation succeeds, 1 if it fails
*/
int SetAttachDetachError_Handlers(PhidgetHandle ch) {
	PhidgetReturnCode prc;
	printf("\n--------------------------------------\n");
	printf("\nSetting OnAttachHandler...\n");
	prc = Phidget_setOnAttachHandler(ch, onAttachHandler, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set Attach Handler: \n\t");
		DisplayError(prc);
		return 1;
	}

	printf("Setting OnDetachHandler...\n");
	prc = Phidget_setOnDetachHandler(ch, onDetachHandler, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set Detach Handler: \n\t");
		DisplayError(prc);
		return 1;
	}

	printf("Setting OnErrorHandler...\n");
	prc = Phidget_setOnErrorHandler(ch, onErrorHandler, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set Error Handler: \n\t");
		DisplayError(prc);
		return 1;
	}

	return 0;
}

/**
* Outputs the VoltageInput's most recently reported voltage.
* Fired when a VoltageInput channel with onVoltageChangeHandler registered meets DataInterval and ChangeTrigger criteria
*
* @param pvih The VoltageInput channel that fired the VoltageChange event
* @param *ctx Context pointer
* @param voltage The reported voltage from the VoltageInput channel
*/
static void CCONV onVoltageChangeHandler(PhidgetVoltageInputHandle pvih, void *ctx, double voltage) {

	printf("[Voltage Event] -> Voltage: %.4f\n", voltage);
}

/**
* Creates a new instance of a VoltageInput channel.
*
* @param pvih Pointer to the PhidgetVoltageInputHandle channel to create
* @return 0 if the operation succeeds, 1 if it fails
*/
static int CreateVoltageInput(PhidgetVoltageInputHandle *pvih) {
	PhidgetReturnCode prc;

	printf("Creating VoltageInput Channel...\n");
	prc = PhidgetVoltageInput_create(pvih);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Creating VoltageInput: \n\t");
		DisplayError(prc);
		return 1;
	}
	return 0;
}

/**
* Sets the event handler for VoltageInput's VoltageChange event
*
* @param pvih The PhidgetVoltageInputHandle channel to add the event to
* @param fptr The callback function to be called when a VoltageChange event is fired
* @return 0 if the operation succeeds, 1 if it fails
*/
static int SetVoltageHandler(PhidgetVoltageInputHandle pvih, PhidgetVoltageInput_OnVoltageChangeCallback fptr) {
	PhidgetReturnCode prc;

	if (fptr) {
		printf("\n--------------------\n"
			"\n  | Voltage change events contain the most recent voltage received from the device.\n"
			"  | The linked VoltageChange function will run as an event at every DataInterval.\n"
			"  | These events will not occur until a change in voltage >= to the set ChangeTrigger has occured.\n"
			"  | DataInterval and ChangeTrigger should initially be set in the device AttachHandler function.\n");
	}
	printf("\n%s OnVoltageChangeHandler...\n", (fptr ? "Setting" : "Clearing"));
	printf("\n--------------------\n");
	prc = PhidgetVoltageInput_setOnVoltageChangeHandler(pvih, fptr, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Setting VoltageChangeHandler: \n\t");
		DisplayError(prc);
		return 1;
	}
	return 0;
}


/**
* Creates, configures, and opens a VoltageInput channel.
* Displays Voltage events for 10 seconds
* Closes out VoltageInput channel
*
* @return 0 if the program exits successfully, 1 if it exits with errors.
*/
int main() {
	PhidgetVoltageInputHandle ch = NULL;

	/*
	 * Allocate a new Phidget Channel object
	 */

	if (CreateVoltageInput(&ch))
		goto error;

	/*
	 * Set matching parameters to specify which channel to open
	 */

	if (SetSerialNumber((PhidgetHandle)ch))
		goto error;

	if (SetVINTProperties((PhidgetHandle)ch))
		goto error;

	if (SetChannel((PhidgetHandle)ch))
		goto error;

	if (SetupNetwork((PhidgetHandle)ch))
		goto error;

	/*
	 * Add event handlers before calling open so that no events are missed.
	 */

	if (SetAttachDetachError_Handlers((PhidgetHandle)ch))
		goto error;

	if (SetVoltageHandler(ch, onVoltageChangeHandler))
		goto error;

	/*
	 * Open the channel with a timeout
	 */

	if (OpenPhidgetChannel_waitForAttach((PhidgetHandle)ch, 5000))
		goto error;

	printf("Sampling data for 10 seconds...\n");

	Sleep(10000);

	/*
	 * Perform clean up and exit
	 */

	if (SetVoltageHandler(ch, NULL))
		goto error;

	printf("\nDone Sampling...\n");

	printf("Cleaning up...\n");
	ClosePhidgetChannel((PhidgetHandle)ch);
	PhidgetVoltageInput_delete(&ch);
	printf("\nExiting...\n");
	printf("Press ENTER to end program.\n");
	getchar();

	return 0;

error:
	PhidgetVoltageInput_delete(&ch);
	fprintf(stderr, "\nExiting with error(s)...\n");
	printf("Press ENTER to end program.\n");
	getchar();

	return 1;
}
